@@ -1,13 +1,12 @@
-pub(crate) use self::event::*;
 use self::input_buffer::*;
 use self::message::*;
 use crate::{
     input::FrameInput,
     time_sync::{TimeSync, UnixMillis},
-    BackrollConfig, Frame, NetworkStats, TaskPool,
+    Config, Frame, NetworkStats, TaskPool,
 };
 use async_channel::TrySendError;
-use backroll_transport::Peer;
+use backroll_transport::Peer as TransportPeer;
 use bincode::config::Options;
 use futures::FutureExt;
 use futures_timer::Delay;
@@ -17,10 +16,11 @@ use serde::{Deserialize, Serialize};
 use std::num::Wrapping;
 use std::sync::Arc;
 use std::time::Duration;
-use tracing::{debug, error, info};
+use tracing::{debug, error};
 
 pub(crate) use event::Event;
 
+mod bitfield;
 mod compression;
 mod event;
 mod input_buffer;
@@ -83,27 +83,23 @@ impl PeerState {
     }
 
     pub fn is_running(&self) -> bool {
-        match self {
-            Self::Running { .. } => true,
-            Self::Interrupted { .. } => true,
-            _ => false,
+        matches!(self, Self::Running { .. } | Self::Interrupted { .. })
+    }
+
+    fn create_sync_request(&self) -> SyncRequest {
+        if let PeerState::Connecting { random, .. } | PeerState::Syncing { random, .. } = self {
+            SyncRequest { random: *random }
+        } else {
+            panic!("Sending sync request while not syncing.")
         }
     }
 
     pub fn is_disconnected(&self) -> bool {
-        if let Self::Disconnected = self {
-            true
-        } else {
-            false
-        }
+        matches!(self, Self::Disconnected)
     }
 
     pub fn is_interrupted(&self) -> bool {
-        if let Self::Interrupted { .. } = self {
-            true
-        } else {
-            false
-        }
+        matches!(self, Self::Interrupted { .. })
     }
 
     pub fn start_syncing(&mut self, round_trips: u8) {
@@ -154,19 +150,19 @@ struct PeerStats {
 }
 
 #[derive(Clone)]
-pub(crate) struct BackrollPeerConfig {
-    pub peer: Peer,
+pub(crate) struct PeerConfig {
+    pub peer: TransportPeer,
     pub disconnect_timeout: Duration,
     pub disconnect_notify_start: Duration,
     pub task_pool: TaskPool,
 }
 
-pub(crate) struct BackrollPeer<T>
+pub(crate) struct Peer<T>
 where
-    T: BackrollConfig,
+    T: Config,
 {
     queue: usize,
-    config: BackrollPeerConfig,
+    config: PeerConfig,
     timesync: TimeSync<T::Input>,
     state: Arc<RwLock<PeerState>>,
 
@@ -182,7 +178,7 @@ where
     events: async_channel::Sender<Event<T::Input>>,
 }
 
-impl<T: BackrollConfig> Clone for BackrollPeer<T> {
+impl<T: Config> Clone for Peer<T> {
     fn clone(&self) -> Self {
         Self {
             queue: self.queue,
@@ -204,15 +200,19 @@ impl<T: BackrollConfig> Clone for BackrollPeer<T> {
     }
 }
 
-impl<T: BackrollConfig> BackrollPeer<T> {
+impl<T: Config> Peer<T> {
     pub fn new(
         queue: usize,
-        config: BackrollPeerConfig,
+        config: PeerConfig,
         local_connect_status: Arc<[RwLock<ConnectionStatus>]>,
     ) -> (Self, async_channel::Receiver<Event<T::Input>>) {
         let (deserialize_send, message_in) = async_channel::unbounded::<Message>();
         let (message_out, serialize_recv) = async_channel::unbounded::<MessageData>();
         let (events, events_rx) = async_channel::unbounded();
+        let peer_connect_status = local_connect_status
+            .iter()
+            .map(|status| status.read().clone())
+            .collect();
         let task_pool = config.task_pool.clone();
 
         let peer = Self {
@@ -223,7 +223,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
 
             stats: Default::default(),
             local_connect_status,
-            peer_connect_status: Default::default(),
+            peer_connect_status,
 
             input_encoder: Default::default(),
             input_decoder: Default::default(),
@@ -240,9 +240,6 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         task_pool
             .spawn(peer.clone().deserialize_incoming(deserialize_send))
             .detach();
-        task_pool
-            .spawn(peer.clone().update_network_stats(NETWORK_STATS_INTERVAL))
-            .detach();
         task_pool.spawn(peer.clone().run()).detach();
 
         (peer, events_rx)
@@ -332,12 +329,6 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         })
     }
 
-    pub fn send_input_ack(&self) -> Result<(), PeerError> {
-        self.send(InputAck {
-            ack_frame: self.input_decoder.last_decoded_frame(),
-        })
-    }
-
     async fn heartbeat(self, interval: Duration) {
         while let Ok(()) = self.send(MessageData::KeepAlive) {
             debug!("Sent keep alive packet");
@@ -415,7 +406,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
                         let mut state = self.state.write();
                         if !state.is_interrupted() && (last_recv_time + notify_start < now) {
                             state.interrupt();
-                            info!("Endpoint has stopped receiving packets for {} ms.  Sending notification.",
+                            debug!("Endpoint has stopped receiving packets for {} ms.  Sending notification.",
                                   notify_start.as_millis());
                             self.push_event(Event::<T::Input>::NetworkInterrupted {
                                 disconnect_timeout: timeout - notify_start
@@ -424,7 +415,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
                     }
 
                     if last_recv_time + timeout < now {
-                        info!(
+                        debug!(
                             "Endpoint has stopped receiving packets for {} ms. Disconnecting.",
                             timeout.as_millis()
                         );
@@ -438,7 +429,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
     }
 
     fn poll(&mut self) -> Result<(), PeerError> {
-        let state = self.state.write();
+        let state = self.state.read();
         let next_interval = match *state {
             PeerState::Connecting { .. } => SYNC_FIRST_RETRY_INTERVAL,
             PeerState::Syncing { .. } => SYNC_RETRY_INTERVAL,
@@ -447,12 +438,16 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         let now = UnixMillis::now();
         if let Some(last_send_time) = self.stats.read().last_send_time {
             if last_send_time + next_interval < now {
-                info!(
+                debug!(
                     "No luck syncing after {:?} ms... Re-queueing sync packet.",
                     next_interval
                 );
-                self.send_sync_request()?;
+                self.send(state.create_sync_request())?;
             }
+        } else {
+            // If we have not sent anything yet, kick off the connection with a
+            // sync request.
+            self.send(state.create_sync_request())?;
         }
 
         Ok(())
@@ -527,7 +522,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
                 // filter out out-of-order packets
                 let skipped = seq - next_recv_seq;
                 if skipped > MAX_SEQ_DISTANCE {
-                    info!(
+                    debug!(
                         "dropping out of order packet (seq: {}, last seq: {})",
                         seq, next_recv_seq
                     );
@@ -564,14 +559,6 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         }
     }
 
-    fn send_sync_request(&self) -> Result<(), PeerError> {
-        if let PeerState::Syncing { random, .. } = *self.state.read() {
-            self.send(SyncRequest { random })
-        } else {
-            panic!("Sending sync request while not syncing.")
-        }
-    }
-
     async fn update_network_stats(self, interval: Duration) {
         let mut start_time: Option<UnixMillis> = None;
 
@@ -617,7 +604,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         let SyncRequest { random } = data;
         if let PeerState::Running { remote_magic } = *self.state.read() {
             if magic != remote_magic {
-                info!(
+                debug!(
                     "Ignoring sync request from unknown endpoint ({} != {:?}).",
                     magic, remote_magic
                 );
@@ -632,7 +619,7 @@ impl<T: BackrollConfig> BackrollPeer<T> {
         let mut state = self.state.write();
         if let Some(random) = state.random() {
             if data.random != random {
-                info!("sync reply {} != {}.  Keep looking...", data.random, random);
+                debug!("sync reply {} != {}.  Keep looking...", data.random, random);
                 return Err(PeerError::InvalidMessage);
             }
         }
@@ -641,22 +628,22 @@ impl<T: BackrollConfig> BackrollPeer<T> {
             PeerState::Connecting { .. } => {
                 self.push_event(Event::<T::Input>::Connected)?;
                 state.start_syncing(NUM_SYNC_PACKETS);
+                self.send(state.create_sync_request())?;
                 Ok(())
             }
             PeerState::Syncing {
                 ref mut roundtrips_remaining,
                 ..
             } => {
-                info!(
+                debug!(
                     "Checking sync state ({} round trips remaining).",
                     *roundtrips_remaining
                 );
                 debug_assert!(*roundtrips_remaining > 0);
                 *roundtrips_remaining -= 1;
                 if *roundtrips_remaining == 0 {
-                    info!("Synchronized queue {}!", self.queue);
+                    debug!("Synchronized queue {}!", self.queue);
                     self.push_event(Event::<T::Input>::Synchronized)?;
-                    self.input_decoder.reset();
                     self.stats.write().last_input_packet_recv_time = UnixMillis::now();
                     *state = PeerState::Running {
                         remote_magic: magic,
@@ -676,19 +663,22 @@ impl<T: BackrollConfig> BackrollPeer<T> {
                     task_pool
                         .spawn(self.clone().resend_inputs(QUALITY_REPORT_INTERVAL))
                         .detach();
+                    task_pool
+                        .spawn(self.clone().update_network_stats(NETWORK_STATS_INTERVAL))
+                        .detach();
                 } else {
                     self.push_event(Event::<T::Input>::Synchronizing {
                         total: NUM_SYNC_PACKETS,
                         count: NUM_SYNC_PACKETS - *roundtrips_remaining as u8,
                     })?;
-                    self.send_sync_request()?;
+                    self.send(state.create_sync_request())?;
                 }
                 Ok(())
             }
-            PeerState::Running { remote_magic } if magic == remote_magic => return Ok(()),
+            PeerState::Running { remote_magic } if magic == remote_magic => Ok(()),
             _ => {
-                info!("Ignoring SyncReply while not syncing.");
-                return Err(PeerError::InvalidMessage);
+                debug!("Ignoring SyncReply while not syncing.");
+                Err(PeerError::InvalidMessage)
             }
         }
     }
@@ -722,6 +712,9 @@ impl<T: BackrollConfig> BackrollPeer<T> {
                 if !inputs.is_empty() {
                     self.push_event(Event::<T::Input>::Inputs(inputs))?;
                     self.stats.write().last_input_packet_recv_time = UnixMillis::now();
+                    self.send(InputAck {
+                        ack_frame: self.input_decoder.last_decoded_frame(),
+                    })?;
                 }
             }
             Err(err) => {
